#include "mainwindow.h"
#include "dbio.h"
#include "displayutil.h"
#include "brlcad/ConstDatabase.h"
#include "brlcad/Object.h"
#include "brlcad/Combination.h"
#include <QApplication>
#include <QMenuBar>
#include <QAction>
#include <QKeySequence>
#include <QFileDialog>
#include <QStringList>
#include <QListWidget>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QMessageBox>
#include <QImage>
#include <QLabel>
#include <QPixmap>
#include <QSplitter>
#include <QPushButton>
#include <QLineEdit>
#include <string>
#include <iostream>


void
MainWindow::setDatabaseName(const char *databaseName)
{
	std::string title = "";
	if (databaseName) {
		title += databaseName;
		title += " - BRLCAD-GUI";
	} else {
		title += "BRLCAD-GUI";
	}
	setWindowTitle(title.c_str());
}


void
MainWindow::openFileDialog()
{
	QFileDialog *dialog = new QFileDialog();
	QStringList *filters = new QStringList();
	*filters << "BRL-CAD Database Files (*.g)"
		    << "All files (*)";
	dialog->setNameFilters(*filters);
	dialog->setFileMode(QFileDialog::AnyFile);
	if (dialog->exec()) {
    	QStringList fileNames = dialog->selectedFiles();
    	std::string fileName = fileNames[0].toStdString();
    	const char *fileNamePtr = fileName.c_str();
    	if (database)
    	 	closeDatabase();
    	loadDatabase(fileNamePtr);
	}
}


void
MainWindow::loadDatabase(const char *fileName)
{
	if (DatabaseUtils::openDatabase(database, fileName) == 0) {
		setDatabaseName(database->Title());
		std::vector<const char*> objects = loadTopLevelObjects();
		setInterfaceVisible(true);
    }
}

void
MainWindow::closeDatabase()
{
	setDatabaseName(NULL);
	setInterfaceVisible(false);
}


void
MainWindow::setInterfaceVisible(bool visible)
{
	if (visible) {
		actionClose->setVisible(true);

	} else {
		actionClose->setVisible(false);
		objectListWidget->clear();
	}
}


void
MainWindow::quit()
{
	QApplication::instance()->quit();
}


void
MainWindow::initMenu()
{
	QMenuBar *menuBar = this->menuBar();

	actionOpen = new QAction(tr("Open Database"), this);
	actionOpen->setShortcut(QKeySequence::Open);
	connect(actionOpen, &QAction::triggered, this, &MainWindow::openFileDialog);

	actionClose = new QAction(tr("Close Database"), this);
	actionClose->setShortcut(QKeySequence::Close);
	connect(actionClose, &QAction::triggered, this, &MainWindow::closeDatabase);
	actionClose->setVisible(false);

	actionQuit = new QAction(tr("Quit"), this);
	actionQuit->setShortcut(QKeySequence::Quit);
	connect(actionQuit, &QAction::triggered, this, &MainWindow::quit);

	QMenu *fileMenu = menuBar->addMenu(tr("&File"));
	fileMenu->addAction(actionOpen);
	fileMenu->addSeparator();
	fileMenu->addAction(actionClose);
	fileMenu->addSeparator();
	fileMenu->addAction(actionQuit);
}


void
MainWindow::initUI() {
	objectListWidget = new QListWidget(this);
	objectListWidget->setCurrentRow(0);
	connect(objectListWidget, &QListWidget::itemDoubleClicked, 
            this, &MainWindow::onObjectDoubleClicked);
	connect(objectListWidget, &QListWidget::itemClicked, 
            this, &MainWindow::onObjectClicked);

	rayTraceButton = new QPushButton(this);
	rayTraceButton->setText(tr("Raytrace"));
	connect(rayTraceButton, &QPushButton::clicked, this, &MainWindow::rayTraceSelectedListItem);

	QLabel *objectListLabel = new QLabel(this);
	objectListLabel->setText("<h3>" + tr("Objects") + "</h3>");

	QWidget *centerWidget = new QWidget(this);
	QHBoxLayout *mainDisplayLayout= new QHBoxLayout(centerWidget);
	mainDisplayWidget = new QLabel(this);
	mainDisplayLayout->addStretch();
	mainDisplayLayout->addWidget(mainDisplayWidget);
	mainDisplayLayout->addStretch();
	centerWidget->setStyleSheet("background-color: #44000000");
	mainDisplayWidget->setStyleSheet("background-color: none;");

	QSplitter *centralContainer = new QSplitter(this);
	QWidget *objectListContainerWrapper = new QWidget(this);
	QVBoxLayout *objectListContainer = new QVBoxLayout(objectListContainerWrapper);

	objectListContainer->addWidget(objectListLabel);
	objectListContainer->addSpacing(10);
	objectListContainer->addWidget(objectListWidget);

	QHBoxLayout *cameraPositionLayout = new QHBoxLayout();
	QLabel *cameraPositionLabel = new QLabel(this);
	cameraXEdit = new QLineEdit(this);
	cameraYEdit = new QLineEdit(this);
	cameraZEdit = new QLineEdit(this);
	cameraXEdit->setText("0");
	cameraYEdit->setText("0");
	cameraZEdit->setText("0");
	cameraPositionLabel->setText(tr("Position"));
	cameraPositionLabel->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
	cameraPositionLayout->addWidget(cameraPositionLabel);
	cameraPositionLayout->addWidget(cameraXEdit);
	cameraPositionLayout->addWidget(cameraYEdit);
	cameraPositionLayout->addWidget(cameraZEdit);

	QHBoxLayout *cameraDirectionLayout = new QHBoxLayout();
	QLabel *cameraDirectionLabel = new QLabel(this);
	cameraDirXEdit = new QLineEdit(this);
	cameraDirYEdit = new QLineEdit(this);
	cameraDirZEdit = new QLineEdit(this);
	cameraDirXEdit->setText("0");
	cameraDirYEdit->setText("0");
	cameraDirZEdit->setText("0");
	cameraDirectionLabel->setText(tr("Direction"));
	cameraDirectionLabel->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
	cameraDirectionLayout->addWidget(cameraDirectionLabel);
	cameraDirectionLayout->addWidget(cameraDirXEdit);
	cameraDirectionLayout->addWidget(cameraDirYEdit);
	cameraDirectionLayout->addWidget(cameraDirZEdit);

	QHBoxLayout *cameraRotationLayout = new QHBoxLayout();
	QLabel *cameraRotationLabel = new QLabel(this);
	cameraRotationEdit = new QLineEdit(this);
	cameraRotationEdit->setText("0");
	cameraRotationLabel->setText(tr("Rotation"));
	cameraRotationLabel->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
	cameraRotationLayout->addWidget(cameraRotationLabel);
	cameraRotationLayout->addWidget(cameraRotationEdit);

	QHBoxLayout *cameraSizeLayout = new QHBoxLayout();
	QLabel *cameraSizeLabel = new QLabel(this);
	cameraWidthEdit = new QLineEdit(this);
	cameraHeightEdit = new QLineEdit(this);
	cameraWidthEdit->setText("0");
	cameraHeightEdit->setText("0");
	cameraSizeLabel->setText(tr("Size"));
	cameraSizeLabel->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
	cameraSizeLayout->addWidget(cameraSizeLabel);
	cameraSizeLayout->addWidget(cameraWidthEdit);
	cameraSizeLayout->addWidget(cameraHeightEdit);

	QHBoxLayout *rayTracePixelLayout = new QHBoxLayout();
	QLabel *rayTracePixelLabel = new QLabel(this);
	rayTracePixelEdit = new QLineEdit(this);
	rayTracePixelEdit->setText("1");
	rayTracePixelLabel->setText(tr("Pixels/ray (less=clearer)"));
	rayTracePixelLabel->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
	rayTracePixelLayout->addWidget(rayTracePixelLabel);
	rayTracePixelLayout->addWidget(rayTracePixelEdit);

	QLabel *cameraControlLabel = new QLabel(this);
	cameraControlLabel->setText("<h3>" + tr("Camera settings") + "</h3>");
	QLabel *rayTraceControlLabel = new QLabel(this);
	rayTraceControlLabel->setText("<h3>" + tr("Raytrace settings") + "</h3>");

	QWidget *cameraControlWidget = new QWidget(this);
	QVBoxLayout *cameraControlLayout = new QVBoxLayout(cameraControlWidget);
	cameraControlLayout->addWidget(cameraControlLabel);
	cameraControlLayout->addSpacing(10);
	cameraControlLayout->addLayout(cameraPositionLayout);
	cameraControlLayout->addLayout(cameraDirectionLayout);
	cameraControlLayout->addLayout(cameraRotationLayout);
	cameraControlLayout->addLayout(cameraSizeLayout);
	cameraControlLayout->addSpacing(20);
	cameraControlLayout->addWidget(rayTraceControlLabel);
	cameraControlLayout->addSpacing(10);
	cameraControlLayout->addLayout(rayTracePixelLayout);
	cameraControlLayout->addWidget(rayTraceButton);
	cameraControlLayout->addSpacing(10);
	cameraControlLayout->addStretch(1);

	centralContainer->setOrientation(Qt::Horizontal);
	centralContainer->addWidget(objectListContainerWrapper);
	centralContainer->addWidget(centerWidget);
	centralContainer->addWidget(cameraControlWidget);
	centralContainer->setStretchFactor(0, 1);
	centralContainer->setStretchFactor(1, 1);
	centralContainer->setStretchFactor(2, 1);

	cameraControlWidget->setMaximumWidth(300);
	cameraControlWidget->setMinimumWidth(175);
	objectListContainerWrapper->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);

	setCentralWidget(centralContainer);
}


void
MainWindow::updateCamera(Camera *camera)
{
	double *pos = camera->position;
	double *dir = camera->direction;
	double *size = camera->size;
	pos[0] = cameraXEdit->text().toDouble();
	pos[1] = cameraYEdit->text().toDouble();
	pos[2] = cameraZEdit->text().toDouble();
	dir[0] = cameraDirXEdit->text().toDouble();
	dir[1] = cameraDirYEdit->text().toDouble();
	dir[2] = cameraDirZEdit->text().toDouble();
	camera->rotation = cameraRotationEdit->text().toDouble();
	size[0] = cameraWidthEdit->text().toDouble();
	size[1] = cameraHeightEdit->text().toDouble();
}


void
MainWindow::updateCamera()
{
	updateCamera(mainCamera);
}


void
MainWindow::updateCameraSettings(const Camera *camera)
{
	double *pos = camera->position;
	double *dir = camera->direction;
	double rot = camera->rotation;
	double *size = camera->size;
	cameraXEdit->setText(QString::number(pos[0]));
	cameraYEdit->setText(QString::number(pos[1]));
	cameraZEdit->setText(QString::number(pos[2]));
	cameraDirXEdit->setText(QString::number(dir[0]));
	cameraDirYEdit->setText(QString::number(dir[1]));
	cameraDirZEdit->setText(QString::number(dir[2]));
	cameraRotationEdit->setText(QString::number(rot));
	cameraWidthEdit->setText(QString::number(size[0]));
	cameraHeightEdit->setText(QString::number(size[1]));
}


void
MainWindow::updateCameraSettings()
{
	updateCameraSettings(mainCamera);
}


void
MainWindow::rayTraceSelectedListItem()
{
	if (objectListWidget->currentItem()) {
		std::string objectName = objectListWidget->currentItem()->text().toStdString();
		const char *objectNamePtr = objectName.c_str();
		rayTrace(objectNamePtr);
	}
}


void
MainWindow::rayTrace(const char *object)
{
	updateCamera(mainCamera);
	int pixelPerRay = rayTracePixelEdit->text().toInt();
	QPixmap *pixmap = DisplayUtils::rayTrace(database, object, mainCamera, pixelPerRay);
	mainDisplayWidget->setPixmap(*pixmap);
}


void
MainWindow::onObjectClicked(QListWidgetItem *listItem)
{
	rayTraceButton->setText(tr("Raytrace") + " \"" + listItem->text() + "\"");
}


void
MainWindow::onObjectDoubleClicked(QListWidgetItem *listItem)
{
	std::string objectName = listItem->text().toStdString();
	const char *objectNamePtr = objectName.c_str();
	database->UnSelectAll();
	database->Select(objectNamePtr);
	BRLCAD::Vector3D boxMinima = database->BoundingBoxMinima();
	BRLCAD::Vector3D boxMaxima = database->BoundingBoxMaxima();
	const double *minimaCoordinates = boxMinima.coordinates;
	const double *maximaCoordinates = boxMaxima.coordinates;
	QString windowTitle = QString("Object \"%1\"").arg(objectNamePtr);
	QMessageBox *messageBox = new QMessageBox(this);
	QString minimaText = "BoundingBoxMinima: (%1, %2, %3)";
	QString maximaText = "BoundingBoxMaxima: (%1, %2, %3)";

	messageBox->setWindowTitle(windowTitle);
	minimaText = minimaText.arg(minimaCoordinates[0]).arg(minimaCoordinates[1]).arg(minimaCoordinates[2]);
	maximaText = maximaText.arg(maximaCoordinates[0]).arg(maximaCoordinates[1]).arg(maximaCoordinates[2]);
	messageBox->setText(minimaText+"\n"+maximaText);
	messageBox->exec();
}


std::vector<const char*>
MainWindow::loadTopLevelObjects()
{
	Camera *camera = new Camera();
	std::vector<const char*> objects;

	BRLCAD::ConstDatabase::TopObjectIterator iter = database->FirstTopObject();
	while (iter.Good())  {
		const char *objectName = iter.Name();
		objects.push_back(objectName);
		++iter;
	}
	for (int i = 0; i < objects.size(); i++)
		if (strcmp(objects[i],"_GLOBAL") != 0)
			objectListWidget->addItem(QString(objects[i]));
	return objects;
} 


void
MainWindow::debug()
{
	double *test = new double[2]{0.6, 0.5};
	double *test1 = new double[2]{1, 0.2};
	double *test3D = DisplayUtils::mapTo3D(test, mainCamera);
	double *testRet2D = DisplayUtils::mapTo2D(test3D, mainCamera);
	std::cout << test[0] << " " << test[1] << std::endl;
	std::cout << testRet2D[0] << " " << testRet2D[1] << std::endl;
	double *test3D1 = DisplayUtils::mapTo3D(test1, mainCamera);
	double *testRet2D1 = DisplayUtils::mapTo2D(test3D1, mainCamera);
	std::cout << test1[0] << " " << test1[1] << std::endl;
	std::cout << testRet2D1[0] << " " << testRet2D1[1] << std::endl;
}


MainWindow::MainWindow(QWidget *parent): QMainWindow(parent)
{
	database = new BRLCAD::ConstDatabase();
	initMenu();
	initUI();
	resize(1280, 720);
	mainCamera = new Camera();
	// debug();
	updateCameraSettings();
}
