#include "brlcad/cicommon.h"
#include "brlcad/ConstDatabase.h"
#include "displayutil.h"
#include <QImage>
#include <QPixmap>
#include <iostream>
#include <cmath>


Camera::Camera()
{
	position = new double[3]{100, 0, 0};
	size = new double[2]{200, 200};
	direction = new double[3]{-1, 0, 0};
	rotation = 0;
}


class Callback: public BRLCAD::ConstDatabase::HitCallback {
public:
	double red;
	double blue;
	double green;
	BRLCAD::Vector3D surfaceNormalIn;
	bool hasHit = false;

	bool operator()(const BRLCAD::ConstDatabase::Hit& hit) throw() {
		if (hit.HasColor()) {
			red = hit.Red();
			green = hit.Green();
			blue = hit.Blue();
			surfaceNormalIn = hit.SurfaceNormalIn();
			hasHit = true;
			return true;
		}
		return false;
	}
};


double
DisplayUtils::mag(double *vec)
{
	return std::sqrt(std::pow(vec[0], 2) + std::pow(vec[1], 2) + std::pow(vec[2], 2)); 
}


double
DisplayUtils::mag(double x, double y, double z)
{
	return std::sqrt(std::pow(x, 2) + std::pow(y, 2) + std::pow(z, 2)); 
}


double
DisplayUtils::mag(double x, double y)
{
	return std::sqrt(std::pow(x, 2) + std::pow(y, 2)); 
}


double
DisplayUtils::dot(double *x, double *y) {
	return x[0]*y[0] + x[1]*y[1] + x[2]*y[2];
}


double *
DisplayUtils::normalize(double *x) {
	double magnitude = mag(x);
	if (magnitude == 0) {
		return new double[3]{0, 0, 0};
	}
	return new double[3]{x[0]/magnitude, x[1]/magnitude, x[2]/magnitude};
}


double *
DisplayUtils::reverse(double *x) {
	return new double[3]{-x[0], -x[1], -x[2]};
}


double
DisplayUtils::angleBetween(double *a, double *b)
{
	/* A.B = |A||B|costheta -> theta = acos(A.B/|A||B|)*/
	double ab = a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
	return std::acos(ab/mag(a)/mag(b));
}


double
DisplayUtils::angle(double x, double y) {
	if (x == 0) {
		if (y >= 0) {
			return M_PI/2;
		} else {
			return -M_PI/2;
		}
	}
	double baseAngle = atan(std::abs(y)/std::abs(x));
	if (x > 0) {
		if (y >= 0) {
			return baseAngle;
		} else {
			return -baseAngle;
		}
	} else {
		if (y >= 0) {
			return M_PI - baseAngle;
		} else {
			return M_PI + baseAngle;
		}
	}
}


/*
 * Project rays parallel to the camera's facing direction from the camera's plane onto the object.
*/
QPixmap *
DisplayUtils::rayTrace(BRLCAD::ConstDatabase *database, const char *object, const Camera *camera, int pixelPerRay)
{

	using namespace BRLCAD;
	database->UnSelectAll();
	database->Select(object);
	Vector3D boxMinima = database->BoundingBoxMinima();
	Vector3D boxMaxima = database->BoundingBoxMaxima();
	double *min = boxMinima.coordinates;
	double *max = boxMaxima.coordinates;

	double ySize = max[1] - min[1];
	double zSize = max[2] - min[2];

	double *position = camera->position;
	double *size = camera->size;

	int imageSize = 500;
	int rows = imageSize/pixelPerRay;
	int cols = imageSize/pixelPerRay;
	int imageWidth = imageSize;
	int imageHeight = imageSize;

	/* Start from the bottom left of the camera */
	double *bl = new double[2]{0, 0};
	double *tl = new double[2]{0, 1};
	double *br = new double[2]{1, 0};

	double *bottomLeft = mapTo3D(bl, camera);
	double *topLeft = mapTo3D(tl, camera);
	double *bottomRight = mapTo3D(br, camera);;

	/* Evenly distribute all rays across camera viewspace */
	double *intervalX = new double[3];
	//size[0]/(cols - 1);
	double *intervalY = new double[3];

	for (int i = 0; i < 3; i++) {
		intervalX[i] = (bottomRight[i] - bottomLeft[i])/(cols - 1);
		intervalY[i] = (topLeft[i] - bottomLeft[i])/(rows - 1);
	}

	QImage *image = new QImage(imageWidth, imageHeight, QImage::Format_RGB32);
	image->fill(qRgba(0, 0, 0, 0));

	for (double x = 0; x < cols; x++) {
		for (double y = 0; y < rows; y++) {
			double *rayOrigin2D = new double[2]{x/cols, y/rows};
			Vector3D *rayOrigin = new Vector3D(mapTo3D(rayOrigin2D, camera));
			Vector3D *rayDirection = new Vector3D(camera->direction);
			Ray3D ray = {*rayOrigin, *rayDirection};
			Callback *callback = new Callback();
			database->ShootRay(ray, *callback, ConstDatabase::StopAfterFirstHit);
			double *location = mapTo2D(rayOrigin->coordinates, camera);

			if (callback->hasHit) {
				double red = callback->red;
				double green = callback->green;
				double blue = callback->blue;
				Vector3D surfaceNormal = callback->surfaceNormalIn;

				double brightness = 0;
				double ambient = 0.1;
				double diffuseWeight = 0.5;
				double specularWeight = 0.5;

				double *dir = DisplayUtils::normalize(rayDirection->coordinates);
				double *norm = DisplayUtils::reverse(DisplayUtils::normalize(surfaceNormal.coordinates));
				double dotProduct = DisplayUtils::dot(dir, norm);

				/* value from 0 to 1 */
				double diffuse = std::max(0.0, dotProduct);

				/* refleced = incidence - 2 normal */
				int n = 4;
				double *reflectedDir = new double[3]{dir[0] - 2*dotProduct*norm[0], dir[1] - 2*dotProduct*norm[1], dir[2] - 2*dotProduct*norm[2]};

				/* value from 0 to 1 */
				double reflectedDotCamDir = DisplayUtils::dot(DisplayUtils::normalize(reflectedDir), DisplayUtils::normalize(DisplayUtils::reverse(camera->direction)));
				reflectedDotCamDir = std::max(0.0, reflectedDotCamDir);
				double specular = std::pow(reflectedDotCamDir, n);
				
				brightness += ambient + diffuse*diffuseWeight;
				red = std::min(red*brightness + specular*specularWeight, 1.0);
				green = std::min(green*brightness + specular*specularWeight, 1.0);
				blue = std::min(blue*brightness + specular*specularWeight, 1.0);

				int pixelX = x/cols*imageWidth;
				int pixelY = imageHeight - y/rows*imageHeight;

				image->setPixel(pixelX, pixelY, QColor((int)(255*red), (int)(255*green), (int)(255*blue)).rgb());
			}
		}
	}
	QPixmap *pixmap = new QPixmap();
	pixmap->convertFromImage(*image, (Qt::ImageConversionFlags) Qt::AutoColor);
	return pixmap;
}


/*
 * Return an array with the x and y values, each between 0 and 1, indicative of their location in the viewport.
*/
double *
DisplayUtils::mapTo2D(double *vec, const Camera *camera)
{
	double *camPos = camera->position;
	double *camDir = camera->direction;
	double *camSize = camera->size;
	double camRot = camera->rotation;

	double *displacement = new double[3]{vec[0] - camPos[0], vec[1] - camPos[1], vec[2] - camPos[2]};
	/* Note, the displacement vector is also the direction vector from the point on the camera. */

	/* We want to calculate the point of the 3D camera from which the ray that corresponds to this point shot. */
	/* If this point is named P, the position (centre) of the camera named C and the given point V, 
	 * The desired vector CP (parallel to direction) is the foot of the perpendicular (F) of the project of CV (displacement) onto the direction vector CD (direction).
	 * CP = FV = CV - CD(CV.CD)/|CD|^2
	*/	

	double CVdotCD = displacement[0]*camDir[0] + displacement[1]*camDir[1] + displacement[2]*camDir[2];
	double magCD = mag(camDir);
	double scaleFactor = CVdotCD/magCD/magCD;
	double *CP = new double[3]{vec[0] - scaleFactor*camDir[0], vec[1] - scaleFactor*camDir[1], vec[2] - scaleFactor*camDir[2]};

	/* Now that we have the vector to the point from the origin of the camera,
	 * we need to rotate it such that the camera is pointing to (0, 1, 0).
	*/

	/* Rotate about Y to rotate Z component into X */
	double angleXZ = angle(camDir[0], camDir[2]);
	double newXMag = mag(camDir[0], camDir[2]);
	double *newDir = new double[3]{newXMag, camDir[1], 0};
	double CP_XZMag = mag(CP[0], CP[2]);
	double CP_XZAngle = angle(CP[0], CP[2]);
	CP = new double[3]{CP_XZMag*std::cos(CP_XZAngle - angleXZ), CP[1], CP_XZMag*std::sin(CP_XZAngle - angleXZ)};

	/* Rotate about Z to rotate X component into Y */
	double angleYX = angle(newDir[1], newDir[0]);
	double CP_YXMag = mag(CP[1], CP[0]);
	double CP_YXAngle = angle(CP[1], CP[0]);
	CP = new double[3]{CP_YXMag*std::sin(CP_YXAngle - angleYX), CP_YXMag*std::cos(CP_YXAngle - angleYX), CP[2]};

	/* Now the camera is theoretically facing (0, 1, 0)
	 * and the coordinates of the point in 2D is simply (X, Z)
	*/
	/* Rotate the point back */
	double magnitude = mag(CP[0], CP[2]);
	double ang = angle(CP[0], CP[2]);
	double camRotRad = camRot/180*M_PI;
	double *point2D = new double[2]{magnitude*cos(ang - camRotRad), magnitude*sin(ang - camRotRad)};
	double *normalized2D = new double[2]{0.5 - point2D[0]/camSize[0], 0.5 - point2D[1]/camSize[1]};
	return normalized2D;
}


/*
 * Returns a 3D vector that corresponds to a possible point that a 2D pixel maps to.
*/
double *
DisplayUtils::mapTo3D(double *initial, const Camera *camera) {
	using namespace std;
	double *camPos = camera->position;
	double *camDir = camera->direction;
	double *camSize = camera->size;
	double camRot = camera->rotation;

	/* This creates a vector that points to the bottom left in 2D
	   which can be used, if the camera is directly facing forwards.
	   In other words, the camera is not rotated about any axis
	*/

	double initialX2D = (initial[0] - 0.5)*camSize[0];
	double initialY2D = (initial[1] - 0.5)*camSize[1];
	double angleInitial = DisplayUtils::angle(initialX2D, initialY2D);
	double r = DisplayUtils::mag(initialX2D, initialY2D);

	/* First we rotate it about the axis parallel to the direction vector of the camera (roll). */
	double rollRad = camRot*M_PI/180;
	angleInitial += rollRad;
	/* Define y axis as "front" */
	double *rolledVec = new double[3]{r*cos(angleInitial), 0, r*sin(angleInitial)};

	/* Then, we need to tilt the camera up or down and minus or plus 90 deg since this is the bottom or top */
	/* As the camera is already panned, the "x" component is actually the sqrt(x^2+y^2) */
	double xyDir = DisplayUtils::mag(camDir[0], camDir[1]);
	double tiltAngle = DisplayUtils::angle(xyDir, camDir[2]);
	/* We obtain the angle to tilt the camera by */
	/* Then we resolve the camera to face the Y direction first */
	/* Then we tilt it, then we pan it back */
	/* When tilt is performed, the XZ camera plane changes
	 * in terms of its Z and Y values while retaining its X */

	/* Tilt the camera, aka changing the length of the outward-
	 * pointing vector and up/downward pointing vector */
	angleInitial = DisplayUtils::angle(rolledVec[1], rolledVec[2]);
	double yzMag = DisplayUtils::mag(rolledVec[1], rolledVec[2]);
	double tiltedZ = yzMag*sin(angleInitial + tiltAngle);
	double tiltedY = yzMag*cos(angleInitial + tiltAngle);
	double *tiltedVec = new double[3]{rolledVec[0], tiltedY, tiltedZ};

	/* Then we rotate it about the z-axis (pan). */
	/* We minus PI/2 because the camera initially points in the position Y direction */
	// double angleFromFront = initial
	double panAngle = DisplayUtils::angle(camDir[0], camDir[1]) - M_PI/2;
	r = DisplayUtils::mag(tiltedVec[0], tiltedVec[1]);
	angleInitial = DisplayUtils::angle(tiltedVec[0], tiltedVec[1]);
	double *pannedVec = new double[3]{r*cos(angleInitial + panAngle), r*sin(angleInitial + panAngle), tiltedVec[2]};

	double *ret = new double[3]{pannedVec[0] + camPos[0], pannedVec[1] + camPos[1], pannedVec[2] + camPos[2]};

	return ret;
}